/* 
 * InCoP - Invisible Camouflage Protocol engine.
 *                                                              
 * Copyright (C) 2011  Luis Campo Giralte 
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 *
 * Written by Luis Campo Giralte <luis.camp0.2009@gmail.com> 2011 
 *
 */

#include "incop.h"
#include "incopdbus.h"
#include "callbacks.h"
#include "genericflow.h"
#include "connection.h"

#define INCOPLOG_CATEGORY_NAME INCOP_INTERFACE
#include "log.h"

static ST_InCoPEngine *_incopEngine = NULL;

static int timeout_checker = 180;

/**
 * INCP_Init - Initialize the main structures of the incop
 */
void INCP_Init() {
	register int i,j;
	ST_Callback *current = NULL;
	ST_Interface *interface = NULL;

	_incopEngine = (ST_InCoPEngine*)g_new0(ST_InCoPEngine,1);

	POLG_Init();
	ICDS_Init();
	_incopEngine->incop_status = INCOP_STATE_STOP;
	_incopEngine->is_pcap_file = FALSE;
	_incopEngine->when_pcap_done_exit = FALSE;
	_incopEngine->pcapfd = 0;
	_incopEngine->pcap = NULL;
	_incopEngine->ttl = 64;
	_incopEngine->src_port = 0;
	_incopEngine->src_address = -1;
	_incopEngine->src_mask_address = 0;
	_incopEngine->dst_port = 0;
	_incopEngine->dst_address = -1;
	_incopEngine->dst_mask_address = 0; 
        _incopEngine->send_messages = 0;
        _incopEngine->receive_messages = 0;
        _incopEngine->decrypt_messages = 0;
	_incopEngine->sockrawfd = PCKT_InitToDevice("lo");
	//_incopEngine->sockrawfd = PCKT_Init();
	_incopEngine->show_generated_payloads = FALSE;

	_incopEngine->source = g_string_new("");
	_incopEngine->bus = ICDS_Connect(INCOP_INTERFACE,(void*)_incopEngine);

	/* Only load the callbacks if dbus is running */ 
	if(_incopEngine->bus != NULL) {
		i = 0;
		interface = &ST_PublicInterfaces[0];
		while(interface->name != NULL) {
			/* Loads the methods first */
			current = (ST_Callback*)&(interface->methods[0]);
			j = 0;
			while((current != NULL)&&(current->name != NULL)) {
				ICDS_AddPublicMethod(interface,current);
				j++;
				current = (ST_Callback*)&(interface->methods[j]);
			}
			j = 0;
			current = (ST_Callback*)&(interface->signals[0]);
			while((current != NULL)&&(current->name != NULL)) {
				ICDS_AddPublicMethod(interface,current);
				j++;
				current = (ST_Callback*)&(interface->signals[j]);
			} 
			j = 0;
			current = (ST_Callback*)&(interface->properties[0]);
			while((current!=NULL)&&(current->name != NULL)){
				ICDS_AddPublicProperty(interface,current);
				j++;
				current = (ST_Callback*)&(interface->properties[j]);
			}
			i++;
			interface = &ST_PublicInterfaces[i];
		}		
	}
	
	PKCX_Init();
	SYIN_Init();

	_incopEngine->protocol = PROT_Init();
	_incopEngine->conn = COMN_Init();
	_incopEngine->flowpool = FLPO_Init();
	_incopEngine->detect = DTTN_Init(); 
	COMN_SetFlowPool(_incopEngine->conn,_incopEngine->flowpool);

#ifdef DEBUG
	LOG(INCOPLOG_PRIORITY_DEBUG,"Initialized engine....");
	LOG(INCOPLOG_PRIORITY_DEBUG,"connection manager (0x%x)",_incopEngine->conn);
	LOG(INCOPLOG_PRIORITY_DEBUG,"flowpool (0x%x)",_incopEngine->flowpool);
#endif
	return;
}

/**
 * INCP_SetSource - Sets the source of the network packets
 *
 * @param source a pcap file or a ethernet device
 */
void INCP_SetSource(char *source){
	g_string_printf(_incopEngine->source,"%s",source);
}


/**
 * INCP_ShowGeneratedPayload - Show the generated packets, usefull for debuging. 
 *
 * @param value true or false 
 */
void INCP_ShowGeneratedPayload(int value){

	_incopEngine->show_generated_payloads = value;
	return;
}

/**
 * INCP_Start - Starts the engine 
 */

void INCP_Start() {

	LOG(INCOPLOG_PRIORITY_INFO,
		"Trying to start the engine, status=%s",incop_states_str[_incopEngine->incop_status]);	
	if(_incopEngine->incop_status == INCOP_STATE_STOP) {
		char errbuf[PCAP_ERRBUF_SIZE];

		_incopEngine->is_pcap_file = FALSE;
		_incopEngine->pcap = pcap_open_live(_incopEngine->source->str, PCAP_ERRBUF_SIZE, 1, -1, errbuf);
		if(_incopEngine->pcap == NULL) {
			_incopEngine->pcap = pcap_open_offline(_incopEngine->source->str,errbuf);
			if(_incopEngine->pcap == NULL) {
				fprintf(stderr, "Could not open device/file \"%s\": %s\n", _incopEngine->source->str, errbuf);
				return;
			}
			_incopEngine->is_pcap_file = TRUE;
		}

		if(pcap_setnonblock(_incopEngine->pcap, 1, errbuf) == 1){
			fprintf(stderr, "Could not set device \"%s\" to non-blocking: %s\n", _incopEngine->source->str,errbuf);
			pcap_close(_incopEngine->pcap);
                	_incopEngine->pcap = NULL;
                	return;
        	}
		_incopEngine->pcapfd = pcap_get_selectable_fd(_incopEngine->pcap);
		_incopEngine->incop_status = INCOP_STATE_RUNNING;
                LOG(INCOPLOG_PRIORITY_INFO,"Starting engine",NULL);
	}
}

/**
 * INCP_Stop - Stops the engine
 */
void INCP_Stop() {
	
	LOG(INCOPLOG_PRIORITY_INFO,
		"Trying to stop the engine, status=%s",incop_states_str[_incopEngine->incop_status]);	
	if(_incopEngine->incop_status == INCOP_STATE_RUNNING) {
		// printf("pcap = 0x%x\n",_incopEngine->pcap);
		//if(_incopEngine->pcap != NULL);
		pcap_close(_incopEngine->pcap);
		_incopEngine->pcap = NULL;
		_incopEngine->pcapfd = -1;
		_incopEngine->incop_status = INCOP_STATE_STOP;
                LOG(INCOPLOG_PRIORITY_INFO,"Stoping engine",NULL);
	}
}

/**
 * INCP_StopAndExit - Stops and exit the incop
 */
void INCP_StopAndExit() {
	INCP_Stop();
	INCP_Destroy();
	exit(0);
}

/**
 * INCP_Destroy - Destroy the ST_InCoPEngine type
 */
void INCP_Destroy() {
	ICDS_Destroy();
	// TODO: the flows stored on the connection manager
	// should be returned to the pools.
	// COMN_ReleaseFlows(_incopEngine->conn);
	COMN_ReleaseFlows(_incopEngine->conn);

	close(_incopEngine->sockrawfd);
	g_string_free(_incopEngine->source,TRUE);
	FLPO_Destroy(_incopEngine->flowpool);
	COMN_Destroy(_incopEngine->conn);
	DTTN_Destroy(_incopEngine->detect);
	PROT_Destroy(_incopEngine->protocol);
	PKCX_Destroy();
	POLG_Destroy();
	g_free(_incopEngine);
	_incopEngine = NULL;
	return;
}

/**
 * INCP_Stats - Show statistics related to the ST_InCoPEngine 
 */

void INCP_Stats() {
        PKDE_PrintfStats();
        FLPO_Stats(_incopEngine->flowpool);
        COMN_Stats(_incopEngine->conn);
	DTTN_Stats(_incopEngine->detect);
	fprintf(stdout,"Messages\n");
	fprintf(stdout,"\tSend: %d\n",_incopEngine->send_messages);
	fprintf(stdout,"\tDecrypted: %d\n",_incopEngine->decrypt_messages);
	fprintf(stdout,"\tReceived: %d\n",_incopEngine->receive_messages);
	return;
}

void INCP_SetExitOnPcap(int value){
	_incopEngine->when_pcap_done_exit = value;
}


/**
 * INCP_ProcessIncomingFlow - Process the input, forward messages to other process from the network
 *
 * @param sig The signature detected on the flow
 * @para 
 */

void INCP_ProcessIncomingFlow(ST_Signature *sig,ST_GenericFlow *f,unsigned char *payload,int len) {
	uint32_t ipsrc = PKCX_GetIPSrcAddr();
	uint32_t masksrc = _incopEngine->src_mask_address;
	uint32_t maskdst = _incopEngine->dst_mask_address;
	uint32_t ipdst = PKCX_GetIPDstAddr();

	if(((ipsrc & masksrc)==(_incopEngine->src_address & masksrc))&&
		((ipdst & maskdst)==(_incopEngine->dst_address & maskdst))){
		// The flow in on the network configured
		uint16_t srcport = PKCX_GetUDPSrcPort();	
		uint16_t dstport = PKCX_GetUDPDstPort();	
		
		// the flow haves the same ports configured or the user
		// dont set the ports for other ports
		if(((srcport==_incopEngine->src_port)&&(dstport==_incopEngine->dst_port))||
			((_incopEngine->src_port == 0)&&(_incopEngine->dst_port == 0))){
			ST_Payload pkt;
			ST_Payload *pkt_recover;

			_incopEngine->decrypt_messages ++;
			pkt.payload = payload;
			pkt.len = len;
#ifdef DEBUG
			LOG(INCOPLOG_PRIORITY_DEBUG,
				"Candidate packet from [%s:%d:%d:%s:%d] flow(0x%x) length(%d)",
				PKCX_GetSrcAddrDotNotation(),
				PKCX_GetSrcPort(),
				17,
				PKCX_GetDstAddrDotNotation(),
				PKCX_GetDstPort(),
                       		f,len);
#endif
	
			pkt_recover = PROT_RecoverPayload(_incopEngine->protocol,&pkt,sig);
			if(pkt_recover) {
				_incopEngine->receive_messages ++;
				LOG(INCOPLOG_PRIORITY_INFO,
					"Message received and decrypted flow(0x%x)msg(%s)",f,pkt_recover->payload);

			}			

		}
	}
	return;
}


/**
 * INCP_Run - Main loop, for manage the packets and the dbus-messages. 
 *
 */
void INCP_Run() {
	ST_GenericFlow *flow;
	ST_Signature *signature;
	register int i;
	int nfds,usepcap,ret,update_timers;
        DBusWatch *local_watches[MAX_WATCHES];
	struct timeval currenttime;
	struct timeval lasttimeouttime;
	struct pcap_pkthdr *header;
	unsigned char *pkt_data;
	struct pollfd local_fds[MAX_WATCHES];

        fprintf(stdout,"%s running on %s machine %s\n",INCOP_ENGINE_NAME,
		SYIN_GetOSName(),SYIN_GetMachineName());
        fprintf(stdout,"\tversion %s\n",SYIN_GetVersionName());

        gettimeofday(&lasttimeouttime,NULL);
	update_timers = 1;
	while (TRUE) {
                nfds = 0;
                usepcap = 0;
                gettimeofday(&currenttime,NULL);

                for (i = 0; i < ICDS_GetTotalActiveDescriptors(); i++) {
                        if (ICDS_GetDescriptorByIndex(i) == 0 ||
                            !dbus_watch_get_enabled(ICDS_GetWatchByIndex(i))) {
                                continue;
                        }

                        local_fds[nfds].fd = ICDS_GetDescriptorByIndex(i); 
                        local_fds[nfds].events = ICDS_GetEventsByIndex(i);
                        local_fds[nfds].revents = 0;
                        local_watches[nfds] = ICDS_GetWatchByIndex(i);
                        nfds++;
                }

                if(_incopEngine->incop_status == INCOP_STATE_RUNNING) {
                        local_fds[nfds].fd = _incopEngine->pcapfd;
                        local_fds[nfds].events = POLLIN|POLLPRI|POLLHUP;
                        local_fds[nfds].revents = 0;
                        usepcap = 1;
                }

                ret = poll(local_fds,nfds+usepcap,-1);
                if (ret <0){
                        //perror("poll");
                        break;
                }

                if((local_fds[nfds].revents & POLLIN)&&(_incopEngine->incop_status == INCOP_STATE_RUNNING)){
                        ret = pcap_next_ex(_incopEngine->pcap,(struct pcap_pkthdr*)&header,(unsigned char*)&pkt_data);
			if(ret < 0) {
                                INCP_Stop();
                                usepcap = 0;
                                if(_incopEngine->is_pcap_file == TRUE){
					fprintf(stdout,"Source analyze done.\n");
					if(_incopEngine->when_pcap_done_exit == TRUE)
                                        	break;
                                }
			}else{
				if(PKDE_Decode(header,pkt_data) == TRUE){
					int segment_size;
					int protocol = PKCX_GetIPProtocol();
			
					if(protocol != IPPROTO_UDP)
						continue;
					uint32_t seq = PKCX_GetSequenceNumber();
					unsigned long hash;
						/* Find a ST_GenericFlow object in order to evaluate it */
					flow = COMN_FindConnection(_incopEngine->conn,
						PKCX_GetIPSrcAddr(),
						PKCX_GetSrcPort(),
						protocol,
						PKCX_GetIPDstAddr(),
						PKCX_GetDstPort(),
						&hash);	
							
					if (flow == NULL) {
						flow = FLPO_GetFlow(_incopEngine->flowpool);
						if (flow != NULL) {
							GEFW_SetFlowId(flow,
								PKCX_GetIPSrcAddr(),
								PKCX_GetSrcPort(),
								protocol,
								PKCX_GetIPDstAddr(),
								PKCX_GetDstPort());	
									
							COMN_InsertConnection(_incopEngine->conn,flow,&hash);
#ifdef DEBUG
							LOG(INCOPLOG_PRIORITY_DEBUG,
								"New connection on Pool [%s:%d:%d:%s:%d] flow(0x%x)",
								PKCX_GetSrcAddrDotNotation(),
								PKCX_GetSrcPort(),
								protocol, 
								PKCX_GetDstAddrDotNotation(),
								PKCX_GetDstPort(),
								flow);
#endif 
							/* Check if the flow allready have a ST_MemorySegment attached */
							GEFW_SetArriveTime(flow,&currenttime);	
						}else{
							//WARNING("No flow pool allocated\n");
							continue;
						}
					}
					// check test/pcapfiles directory
					segment_size = PKCX_GetPayloadLength();
					flow->total_packets++;
					flow->total_bytes += segment_size;

					if(flow->detected == 0){ // The flow is not detected
						if(segment_size > 0 ){ // the packet have payload
							signature = DTTN_MatchsSignatures(_incopEngine->detect,
								flow,
								PKCX_GetPayload(),
								segment_size);
							if(signature) {
                                                        	LOG(INCOPLOG_PRIORITY_INFO,
									"Detecting '%s' on flow [%s:%d:%d:%s:%d] flow(0x%x)",
									signature->name,
									PKCX_GetSrcAddrDotNotation(),
									PKCX_GetSrcPort(),
									protocol,
									PKCX_GetDstAddrDotNotation(),
									PKCX_GetDstPort(),
                                                                	flow);

								INCP_ProcessIncomingFlow(signature,flow,
									PKCX_GetPayload(),segment_size);	
							}
						}
					}
					GEFW_UpdateTime(flow,&currenttime);
				} // end of decode;
			}
                }
		/* updates the flow time every 180 seconds aproximately
		 * in order to avoid sorting without non-sense the flow list timer
		 * Notice that if not dbus messages available on the buss the 
		 * timers never execute.
		 */
		if(lasttimeouttime.tv_sec + timeout_checker < currenttime.tv_sec) {
			if(update_timers) {
				COMN_UpdateTimers(_incopEngine->conn,&currenttime);
				update_timers = 0;
				lasttimeouttime.tv_sec = currenttime.tv_sec;
				lasttimeouttime.tv_usec = currenttime.tv_usec;
			}
		}else 
			update_timers = 1;
               	for (i = 0; i < nfds; i++) {
                        if (local_fds[i].revents) {
                                ICDS_Handler(_incopEngine->bus,local_fds[i].revents, local_watches[i]);
                        }
                }
        }
        return;
}

/**
 * INCP_SetEncryptionKey - Sets the initial rc4 encryption key. 
 *
 * @param key the rc4 key
 *
 */
void INCP_SetEncryptionKey(char *key) {

        LOG(INCOPLOG_PRIORITY_INFO,
                "setting encryption key '%s'",key);
	_incopEngine->protocol->rc4_encryption_key = key;
	return;
}

/**
 * INCP_AddSignature - Adds a signature to the sistem. 
 *
 * @param identifier a number that identify the signature.
 * @param name the name of the signature.
 * @param expresion a regular expresssion.
 * @param head the head of the regular expression.
 * @param tail the tail of the regular expresssion. 
 *
 */
void INCP_AddSignature(int identifier,char *name,char *expression,char *head,int hsize, char *tail,int tsize){
	ST_Signature *sig = SGNT_Init();

        LOG(INCOPLOG_PRIORITY_INFO,
                "adding signature '%s' to the engine",name);

	SGNT_SetValues(sig,identifier,name,expression,head,hsize,tail,tsize);

	DTTN_AddSignature(_incopEngine->detect,sig);	

	return;
}

/**
 * INCP_SetSourcePort - Sets the source port for the generated packets. 
 *
 * @param srcport . 
 *
 */
void INCP_SetSourcePort(int srcport){
	_incopEngine->src_port = srcport;
	return;
}

/**
 * INCP_SetDestinationPort - Sets the destination port for the generated packets. 
 *
 * @param dstport . 
 *
 */
void INCP_SetDestinationPort(int dstport){
	_incopEngine->dst_port = dstport;
	return;
}


void INCP_SetSourceIP(char *ipsrc){
	struct in_addr addr;
	char *mask;
	int ret;

	mask = strchr(ipsrc,'/');
	if(mask == NULL) {
		mask = "0";
	}else{
    		*mask++ = '\0';
	}

	ret = inet_aton(ipsrc,&addr);
	if(ret) {
		_incopEngine->src_address = ntohl(addr.s_addr);
		_incopEngine->src_mask_address  = ~0 << (32 - atoi(mask));
	}
	return;
}

void INCP_SetDestinationIP(char *ipdst){
        struct in_addr addr;
        char *mask;
        int ret;

        mask = strchr(ipdst,'/');
        if(mask == NULL) {
                mask = "0";
        }else{
                *mask++ = '\0';
        }

        ret = inet_aton(ipdst,&addr);
        if(ret) {
                _incopEngine->dst_address = ntohl(addr.s_addr);
                _incopEngine->dst_mask_address  = ~0 << (32 - atoi(mask));
        }
	return;
}

/**
 * INCP_ReportIncident - Receive a message from the dbus and send over the network. 
 *
 * @param message message to report to the network
 *
 */
void INCP_ReportIncident(char *message){
	ST_Available *av = NULL;
	int ret;
	uint32_t ipsrc;
	uint32_t ipdst;
	uint16_t portsrc;
	uint16_t portdst;
	struct in_addr a,b;
        char ipsrc_str[INET_ADDRSTRLEN];
        char ipdst_str[INET_ADDRSTRLEN];

	av = DTTN_GetAvailable(_incopEngine->detect);
	if(av) {
		ST_Payload *payload;

		if(_incopEngine->src_address == -1) { // use the source IP of the flow
			ipsrc = av->sig->ipsrc;
		}else{
			if(_incopEngine->src_mask_address == 0){ // There is a source IP address that can be used
				ipsrc = _incopEngine->src_address;
			}else{
				// TODO: if the user configures a network 195.11.100.0/24
				// a good techique will be generate a random source ip from the
				// corresponding network.
				// this works also for the destination address. 
				ipsrc = av->sig->ipsrc; // use the ip source of the detected flow
			}
		}
		if(_incopEngine->dst_address == -1) { // use the destination IP of the flow
			ipdst = av->sig->ipdst;
		}else{
			if(_incopEngine->dst_mask_address == 0){ // There is a source IP address that can be used
				ipdst = _incopEngine->dst_address;
			}else{
				ipdst = av->sig->ipdst; // use the ip dest of the detected flow
			}
                }
		ipdst = av->sig->ipdst;
		portsrc = av->sig->portsrc;		
		portdst = av->sig->portdst;		
		a.s_addr = ntohl(ipsrc);
		b.s_addr = ntohl(ipdst);

        	inet_ntop(AF_INET, &a, ipsrc_str, INET_ADDRSTRLEN);
        	inet_ntop(AF_INET, &b, ipdst_str, INET_ADDRSTRLEN);

		payload = PROT_GeneratePayload(_incopEngine->protocol,av->sig,message,strlen(message));
		ret = PCKT_Send(_incopEngine->sockrawfd,
			ipsrc,
			ipdst,
			_incopEngine->ttl,
			// TODO: If the source of the traffic is on the same machine as incop
			// the ports can not be the same. In any other case could.
			portsrc +1,
			portdst,
			payload->payload,
			payload->len);
        	LOG(INCOPLOG_PRIORITY_INFO,
                	"Reporting incident over '%s' using [%s:%d:17:%s:%d] %d bytes",av->sig->name,
			ipsrc_str,portsrc,ipdst_str,portdst,ret);
		if(_incopEngine->show_generated_payloads == TRUE){
			PYLD_Printf(payload);
		}
		PYLD_Destroy(payload);
	}else{
        	LOG(INCOPLOG_PRIORITY_INFO,
                	"No traffic avaiable to detect");
	}
	return;
}


void INCP_SetPacketTTL(int ttl){
	_incopEngine->ttl = ttl;
	return;
}
